/*
 * Created on Nov 11, 2004
 */
package org.tekkotsu.ui.storyboard.objects;

import org.eclipse.draw2d.Graphics;
import org.eclipse.draw2d.Label;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.PointList;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.RGB;
import org.tekkotsu.ui.editor.model.StateNodeModel;
import org.tekkotsu.ui.storyboard.model.BlockModel;
import org.tekkotsu.ui.storyboard.views.StoryboardViewer;
import org.tekkotsu.ui.util.ColorGenerator;
import org.tekkotsu.ui.util.Debugger;

/**
 * @author asangpet
 *
 */
public class BlockViewObject extends AbstractViewObject {	
	private BlockModel model;
	private StateNodeModel state;
	private Rectangle oldCross;
	
	public BlockViewObject(StoryboardViewer viewer, BlockModel model, StateNodeModel state) {
		super(viewer);
		setState(state);
		setModel(model);
	}
	
	private void setState(StateNodeModel state) {
		this.state = state;
		Color c = state.getColor();
		setColor(new RGB(c.getRed(), c.getGreen(), c.getBlue()));
		setToolTip(new Label(state.getLabel()));
	}

	private void setModel(BlockModel model) {
		this.model = model;
		final Rectangle EDITOR_BOUNDS = state.getConstraint();

		final int EY = EDITOR_BOUNDS.y;
		final int EW = EDITOR_BOUNDS.width;
		final int EH = EDITOR_BOUNDS.height;
		
		final int X = timeline.getOffsetOfTime(model.getStart());
		final int W = timeline.getOffsetOfTime(model.getEnd()) - X;

		//make box in storyboard at least as long as box in editor
		if( W >= EW || !model.isFinish() ) {
			setConstraint(X,EY,W,EH);
		}
		else {
		    timeline.getMarker().addMarker(model.getStart(),X);
		    timeline.getMarker().addMarker(model.getEnd(),X + EW);
			setConstraint(X,EY,EW,EH);
		}
	}
	
	public void setColor(RGB c) {		
		colReg.put(COLOR_BASE, c);		
		colReg.put(COLOR_FADE, ColorGenerator.fade(c,200));
		colReg.put(COLOR_BORDER, ColorGenerator.fade(c,100));
		updateColors();
	}

	@Override
	public void updateColors() {
		if (getActive()) {
			setBackgroundColor(colReg.get(COLOR_BASE));
			setForegroundColor(ColorGenerator.highContrast(getBackgroundColor()));
		} else {
			setForegroundColor(colReg.get(COLOR_BORDER));
			setBackgroundColor(colReg.get(COLOR_FADE));			
		}
	}

	@Override
	protected void paintFigure(Graphics gc) {		

		final Rectangle bounds = getBounds();
		final Rectangle viewbox = viewer.getVisibleViewport();
		final Rectangle cross = bounds.getCopy().intersect(viewbox);
		
		if( !cross.equals(oldCross) ) {
			oldCross = cross.getCopy();
			forceRepaint();
		}
		
		final int BX = bounds.x;
		final int BY = bounds.y;
		final int BW = bounds.width;
		final int BH = bounds.height;
		final int VIEW_L = viewbox.x - 50;
		final int VIEW_R = VIEW_L + viewbox.width + 50;
		final int CHAR_WIDTH = gc.getFontMetrics().getAverageCharWidth();
		int fontOffset = 0;
		
		if( BX + BW < VIEW_L ) {
			Debugger.printDebug("BLOCK TOO LEFT");
			return;
		}
		else if( BX > VIEW_R ) {
			Debugger.printDebug("BLOCK TOO RIGHT");
			return;
		}
		
		/*
		 * Draw the Node
		 */
		gc.setBackgroundColor(getBackgroundColor());
		gc.setForegroundColor(ColorGenerator.borderColor(getBackgroundColor()));
		
		String shape = state.getShape();
		if (shape.equals(StateNodeModel.P_SHAPE_STYLE_ELLIPSE)) {
		    int hbound = Math.min(BH,BW);
		    gc.fillOval(BX,BY,hbound,BH);
		    gc.fillOval(BX+BW-hbound,BY,hbound,BH);		    
		    gc.drawArc(BX,BY,hbound,BH,90,180);
		    gc.drawArc(BX+BW-hbound,BY,hbound,BH,270,180);		    
		    int xoff = BX+hbound/2;
		    gc.fillRectangle(xoff,BY,BW-hbound,BH);
		    gc.drawLine(xoff,BY,xoff+BW-hbound,BY);
		    gc.drawLine(xoff-2,BY+BH-1,xoff+BW-hbound+2,BY+BH-1);
		} else if (shape.equals(StateNodeModel.P_SHAPE_STYLE_HEXAGON)){
			fontOffset = -5;
		    int hbound = state.getConstraint().width/6;
		    if (2*hbound>BW) hbound = BW/6;

		    PointList plist = new PointList();
			plist.addPoint(new Point(BX+hbound,BY));
			plist.addPoint(new Point(BX+BW-hbound,BY));
			plist.addPoint(new Point(BX+BW-1,BY+BH/2));
			plist.addPoint(new Point(BX+BW-hbound,BY+BH-1));
			plist.addPoint(new Point(BX+hbound,BY+BH-1));
			plist.addPoint(new Point(BX,BY+BH/2));
			gc.fillPolygon(plist);
			gc.drawPolygon(plist);
		}  else if (shape.equals(StateNodeModel.P_SHAPE_STYLE_ROUNDED_RECTANGLE)){
		    Rectangle r = new Rectangle(BX,BY,BW,BH);
		    gc.fillRoundRectangle(r,8,8);		
		    r = new Rectangle(BX,BY,BW-1,BH-1);
			gc.drawRoundRectangle(r,8,8);
		} else {
		    // default style is RECTANGLE
			gc.fillRectangle(BX,BY,BW,BH);		
			gc.drawRectangle(BX,BY,BW-1,BH-1);		    
		}
		
		/*
		 * Draw the Label 
		 */
		gc.setForegroundColor(getForegroundColor());
		String label = state.getLabel();
		if (label==null) label = state.getId();		
		int LABEL_W = CHAR_WIDTH*label.length();
		int LABEL_H = gc.getFontMetrics().getHeight();
		
		if (LABEL_W < cross.width) {
			//label shorter than block width, so center
			gc.drawText(label,cross.x+(cross.width-LABEL_W)/2+fontOffset,cross.y+(cross.height-LABEL_H)/2);
		} else {
			//label longer than block width, so left justify
			gc.drawText(label,cross.x+2,cross.y+(cross.height-LABEL_H)/2);
		}
	}
	/**
	 * @return Returns the model.
	 */
	@Override
	public BlockModel getModel() {
		return model;
	}

	@Override
	public String getID() {
		return state.getId();
	}
}
